from typing import Literal

import jax
from jaxonloader import get_tiny_shakespeare, make
from kira import Kira, Mamba
from kira.model_args import get_kira_args, get_mamba_args
from kira.train import train


def main():
    max_seq_len = 8
    early_stop = 500
    batch_size = 64
    tinyshakespeare = get_tiny_shakespeare()
    train_dataset, test_dataset, vocab_size, encode, decode = tinyshakespeare
    key = jax.random.PRNGKey(100)
    key, subkey = jax.random.split(key)
    train_dataloader, train_index = make(
        train_dataset,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
        key=key,
        jit=True,
    )
    test_dataloader, test_index = make(
        test_dataset,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
        key=subkey,
    )
    n_dims = vocab_size
    n_embd = 64  # 384
    learning_rate = 3e-4
    num_heads = 4  # 6
    query_multihead_dim = num_heads
    kv_multihead_dim = 2
    n_layers = 3  # 6
    max_new_tokens = 200  # noqa

    train_kira(
        train_dataloader,
        train_index,
        n_dims,
        n_embd,
        n_layers,
        max_seq_len,
        num_heads,
        query_multihead_dim,
        kv_multihead_dim,
        learning_rate,
        early_stop,
        kv_interpolation_mode="repeat",
    )


def train_mamba(
    train_dataloader,
    train_index,
    n_dims,
    n_embd,
    n_layers,
    learning_rate,
    early_stop,
    key,
):
    model_args = get_mamba_args(
        n_embd=n_embd, n_dims=n_dims, n_layers=n_layers, d_state=4
    )
    mamba = Mamba(model_args=model_args, key=key)
    key, subkey = jax.random.split(key)
    mamba = train(
        train_dataloader,
        train_index,
        learning_rate,
        mamba,
        subkey,
        early_stop=early_stop,
    )
    return mamba


def train_kira(
    train_dataloader,
    train_index,
    n_dims,
    n_embd,
    n_layers,
    max_seq_len,
    num_heads,
    query_multihead_dim,
    kv_multihead_dim,
    learning_rate,
    early_stop,
    kv_interpolation_mode: Literal["average", "repeat"] = "average",
):
    kira_model_args = get_kira_args(
        n_dims=n_dims,
        n_embd=n_embd,
        n_layers=n_layers,
        max_seq_len=max_seq_len,
        num_heads=num_heads,
        num_query_heads=query_multihead_dim,
        num_kv_heads=kv_multihead_dim,
        width_size=256,
        depth=4,
        key_seed=0,
        kv_interpolation_mode=kv_interpolation_mode,
    )
    key = jax.random.PRNGKey(kira_model_args.key_seed)
    kira = Kira(
        model_args=kira_model_args,
        key=key,
    )
    key, subkey = jax.random.split(key)
    kira = train(
        train_dataloader,
        train_index,
        learning_rate,
        kira,
        early_stop=early_stop,
        key=subkey,
    )
    return kira


if __name__ == "__main__":
    with jax.checking_leaks():
        main()



import functools as ft
from typing import Any, Optional

import equinox as eqx
import jax
import jax.numpy as jnp
import optax
from jaxonloader import Index, JaxonDataLoader, JITJaxonDataLoader
from jaxtyping import Array, Int, PRNGKeyArray, PyTree
from loguru import logger


def train(
    train_dataloader: JaxonDataLoader | JITJaxonDataLoader,
    train_index: Index,
    learning_rate: float,
    model: PyTree,
    key: PRNGKeyArray,
    early_stop: int | None = None,
    wandb_client: Any | None = None,
    log_every: Optional[int] = 100,
) -> PyTree:
    optimizer = optax.adamw(learning_rate=learning_rate)
    opt_state = optimizer.init(eqx.filter(model, eqx.is_inexact_array))
    loss_value = 0
    i = 0
    while it := train_dataloader(train_index):
        x, train_index, done = it
        if done:
            break
        x, y = jnp.split(x, 2, axis=1)
        key, subkey = jax.random.split(key)
        model, opt_state, loss_value = step(
            model, opt_state, x, y, optimizer, key=subkey
        )

        if log_every is not None and i % log_every == 0:
            logger.info(f"Loss: {loss_value}")
        if early_stop is not None and i > early_stop:
            break

        i += 1
    logger.info("Finished training")
    logger.info(f"Final loss: {loss_value}")
    return model


def evaluate(
    test_dataloader: JaxonDataLoader,
    test_index: eqx.nn.State,
    model: PyTree,
):
    loss = 0
    while it := test_dataloader(test_index):
        x, index, done = it
        if done:
            break
        x, y = jnp.split(x, 2, axis=1)
        x = jnp.array(x)
        y = jnp.array(y)
        loss += loss_fn(model, x, y, key=None)

    return loss / len(test_dataloader)


@eqx.filter_jit
def loss_fn(
    model: PyTree,
    x: Int[Array, "batch_size max_seq_len n_dims"],
    labels: Int[Array, "batch_size max_seq_len n_dims"],
    key: Optional[PRNGKeyArray],
) -> Array:
    partial_model = ft.partial(model, state=None, key=key)
    logits = eqx.filter_vmap(partial_model)(x)
    return jnp.mean(optax.softmax_cross_entropy_with_integer_labels(logits, labels))


@eqx.filter_jit
def step(
    model: PyTree,
    opt_state: PyTree,
    x: Array,
    y: Array,
    optimizer: optax.GradientTransformation,
    key: PRNGKeyArray,
) -> tuple[PyTree, PyTree, Any]:
    loss, grads = eqx.filter_value_and_grad(loss_fn)(model, x, y, key)
    updates, opt_state = optimizer.update(grads, opt_state, model)
    model = eqx.apply_updates(model, updates)

    return model, opt_state, loss



# Copied from Meta with minor modifications.

# Copyright (c) Meta Platforms, Inc. and affiliates.
# This software may be used and distributed according to the terms of the
# Llama 2 Community License Agreement.

import os
from logging import getLogger
from typing import List

from sentencepiece import SentencePieceProcessor


logger = getLogger()


class Tokenizer:
    """tokenizing and encoding/decoding text using SentencePiece."""

    def __init__(self, model_path: str):
        """
        Initializes the Tokenizer with a SentencePiece model.

        Args:
            model_path (str): The path to the SentencePiece model file.
        """
        # reload tokenizer
        assert os.path.isfile(model_path), model_path
        self.sp_model = SentencePieceProcessor(model_file=model_path)
        logger.info(f"Reloaded SentencePiece model from {model_path}")

        # BOS / EOS token IDs
        self.n_words: int = self.sp_model.vocab_size()
        self.bos_id: int = self.sp_model.bos_id()
        self.eos_id: int = self.sp_model.eos_id()
        self.pad_id: int = self.sp_model.pad_id()
        logger.info(
            f"#words: {self.n_words} - BOS ID: {self.bos_id} - EOS ID: {self.eos_id}"
        )
        assert self.sp_model.vocab_size() == self.sp_model.get_piece_size()

    def encode(self, s: str, bos: bool, eos: bool) -> List[int]:
        """
        Encodes a string into a list of token IDs.

        Args:
            s (str): The input string to be encoded.
            bos (bool): Whether to prepend the beginning-of-sequence token.
            eos (bool): Whether to append the end-of-sequence token.

        Returns:
            List[int]: A list of token IDs.
        """
        assert isinstance(s, str), "Input must be a string."
        t = self.sp_model.encode(s)
        if bos:
            t = [self.bos_id] + t
        if eos:
            t = t + [self.eos_id]
        return t

    def decode(self, t: List[int]) -> str:
        """
        Decodes a list of token IDs into a string.

        Args:
            t (List[int]): The list of token IDs to be decoded.

        Returns:
            str: The decoded string.
        """
        return self.sp_model.decode(t)



from typing import Optional

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, PRNGKeyArray


internal_rope_embedding_cache: dict[int, Array] = {}


class RotaryPositionalEmbedding(eqx.Module):
    embedding_size: int = eqx.field(static=True)

    def __check_init__(self):
        if self.embedding_size < 0:
            raise ValueError("`embedding_size` must not be negative.")
        if (self.embedding_size % 2) != 0:
            raise ValueError("`embedding_size` must be even.")

    @staticmethod
    def rotate_half(x: Float[Array, "seq_length embedding_size"]):
        d_2 = x.shape[-1] // 2
        return jnp.concatenate([-x[..., d_2:], x[..., :d_2]], axis=-1)

    @staticmethod
    def precompute_freqs_cis(
        embedding_size: int, end: int, theta: float = 10000.0
    ) -> Complex[Array, "end half_of_embedding_size"]:
        freqs = 1.0 / (
            theta
            ** (jnp.arange(0.0, embedding_size, 2)[jnp.newaxis, :] / embedding_size)
        )

        t = jnp.arange(float(end))
        freqs_outer = jnp.outer(t, freqs)
        with jax.numpy_dtype_promotion("standard"):
            freqs_cis = jnp.cos(freqs_outer) + jnp.sin(freqs_outer) * 1j

        return freqs_cis

    def __call__(
        self,
        x: Float[Array, "seq_length embedding_size"],
        *,
        key: Optional[PRNGKeyArray] = None,
    ) -> Float[Array, "seq_length embedding_size"]:
        """**Arguments:**

        - `x`: A JAX array of shape `(seq_length, embedding_size)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(seq_length, embedding_size)`, with the rotary positional
        encoding applied to the input.
        """

        seq_len, embedding_size = x.shape
        if embedding_size != self.embedding_size:
            raise ValueError(
                f"x.shape[-1] must match self.embedding_size, "
                f"but {x.shape[-1]} != {self.embedding_size}"
            )

        with jax.ensure_compile_time_eval():
            if embedding_size in internal_rope_embedding_cache:
                freqs_cis = internal_rope_embedding_cache[embedding_size]
                freqs_cis_seq_len, _ = freqs_cis.shape
                if seq_len > freqs_cis_seq_len:
                    freqs_cis = self.precompute_freqs_cis(embedding_size, seq_len)
                    internal_rope_embedding_cache[embedding_size] = freqs_cis
                else:
                    freqs_cis = freqs_cis[:seq_len]
            else:
                freqs_cis = self.precompute_freqs_cis(embedding_size, seq_len)
                internal_rope_embedding_cache[embedding_size] = freqs_cis

        freqs_real = jnp.tile(freqs_cis.real, (1, 2))
        freqs_imag = jnp.tile(freqs_cis.imag, (1, 2))

        rotate_x = self.rotate_half(x)
        x_rope = (x * freqs_real) + (rotate_x * freqs_imag)
        return x_rope



from functools import partial
from typing import Optional

import equinox as eqx
import jax
from jaxtyping import Array, Int, PRNGKeyArray

from kira.model_args import KiraModelArgs
from kira.modules.mha import MultiheadAttention
from kira.modules.rope_embeddings import RotaryPositionalEmbedding


class Block(eqx.nn.StatefulLayer):
    mha_attention: MultiheadAttention
    rms_norm: eqx.nn.RMSNorm
    feedforward: eqx.nn.MLP
    dropout: eqx.nn.Dropout

    model_args: KiraModelArgs = eqx.field(static=True)

    rope_embeddings: RotaryPositionalEmbedding

    def __init__(
        self,
        model_args: KiraModelArgs,
        *,
        key,
        **kwargs,
    ):
        self.model_args = model_args
        key, *subkeys = jax.random.split(key, 5)
        self.rope_embeddings = RotaryPositionalEmbedding(
            embedding_size=model_args.n_embd,
        )

        self.mha_attention = MultiheadAttention(
            num_heads=model_args.num_heads,
            query_size=model_args.n_embd,
            qk_size=model_args.n_embd,
            key_size=model_args.n_embd,
            value_size=model_args.n_embd,
            vo_size=model_args.n_embd,
            output_size=model_args.n_embd,
            query_multihead_dim=model_args.num_query_heads,
            kv_multihead_dim=model_args.num_kv_heads,
            state_length=model_args.max_seq_len,
            kv_interpolation_mode=model_args.kv_interpolation_mode,
            key=subkeys[0],
        )

        self.rms_norm = eqx.nn.RMSNorm(shape=model_args.n_embd)

        self.feedforward = eqx.nn.MLP(
            model_args.n_embd,
            out_size=model_args.n_embd,
            width_size=model_args.width_size,
            depth=model_args.depth,
            key=subkeys[1],
        )

        self.dropout = eqx.nn.Dropout(p=model_args.p)

    def __call__(
        self,
        x: Int[Array, "max_seq_len input_dim"],
        state: eqx.nn.State | None = None,
        mask: str | None = "causal",
        *,
        key: Optional[PRNGKeyArray],
        **kwargs,
    ):
        def process_heads(query_heads, key_heads, value_heads):
            query_heads = jax.vmap(self.rope_embeddings, in_axes=1, out_axes=1)(
                query_heads
            )
            key_heads = jax.vmap(self.rope_embeddings, in_axes=1, out_axes=1)(key_heads)

            return query_heads, key_heads, value_heads

        mha_partial = partial(
            self.mha_attention,
            process_heads=process_heads,
            query=jax.vmap(self.rms_norm)(x),
            key_=jax.vmap(self.rms_norm)(x),
            value=jax.vmap(self.rms_norm)(x),
            mask=mask,
        )
        if state is not None:
            mha, state = mha_partial(state=state, key=key)
        else:
            mha = mha_partial(key=key)
        x = mha + x
        inference = True if key is None else False
        d_key1 = None
        d_key2 = None
        if not inference and key is not None:
            key, d_key1, d_key2 = jax.random.split(key, 3)
        x = self.dropout(x, key=d_key1, inference=inference)
        ff = jax.vmap(self.feedforward)(jax.vmap(self.rms_norm)(x))
        x = ff + x
        x = self.dropout(x, key=d_key2, inference=inference)
        return x, state


class Kira(eqx.Module):
    model_args: KiraModelArgs = eqx.field(static=True)

    blocks: list[Block]

    input_embedding: eqx.nn.Embedding
    output: eqx.nn.Linear

    rms_norm: eqx.nn.RMSNorm

    def __init__(
        self,
        model_args: KiraModelArgs,
        *,
        key,
        **kwargs,
    ):
        assert model_args.n_layers is not None, "n_layers must be provided"
        self.model_args = model_args
        key, *subkeys = jax.random.split(key, model_args.n_layers + 2)

        self.input_embedding = eqx.nn.Embedding(
            model_args.n_dims, model_args.n_embd, key=subkeys[0]
        )
        self.blocks = [
            Block(
                model_args=model_args,
                key=subkeys[i + 1],
            )
            for i in range(model_args.n_layers)
        ]

        self.rms_norm = eqx.nn.RMSNorm(shape=model_args.n_embd)
        self.output = eqx.nn.Linear(
            model_args.n_embd, model_args.n_dims, key=subkeys[-1]
        )

    def __call__(
        self,
        x: Int[Array, " seq_len"],
        state: eqx.nn.State | None = None,
        mask: str | None = "causal",
        *,
        key: PRNGKeyArray | None = None,
    ):
        x = jax.vmap(self.input_embedding)(x)
        for block in self.blocks:
            x, state = block(x, state=state, mask=mask, key=key)
        x = jax.vmap(self.rms_norm)(x)
        x = jax.vmap(self.output)(x)

        if state is not None:
            return x, state
        else:
            return x



from typing import Callable

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import PyTree


def generate_text(
    model: PyTree,
    max_seq_len: int,
    max_new_tokens: int,
    decode: Callable,
    vocab_size: int,
    print_to_console: bool = True,
    random_key_seed: int = 0,
    state: eqx.nn.State = None,
) -> tuple[list[str], list[int]]:
    jitted_model = eqx.filter_jit(model)
    x = jnp.zeros((max_seq_len,), dtype=jnp.int32)
    key = jax.random.PRNGKey(random_key_seed)
    tokens = []
    decoded_tokens = []
    for _ in range(max_new_tokens):
        key, subkey, model_key = jax.random.split(key, 3)
        if state is None:
            logits = jitted_model(x, state=state, key=model_key)
        else:
            logits, state = jitted_model(x, state=state, key=model_key)
        logits = logits[-1, :]
        probs = jax.nn.softmax(logits, axis=-1)

        next_token = jax.random.choice(
            subkey,
            jnp.arange(len(probs)),
            p=probs,
        )
        next_token = jnp.array(next_token, dtype=jnp.int32).reshape((1,))
        next_token = min(next_token.item(), vocab_size - 1)

        if print_to_console:
            print(decode([next_token]), end="")

        tokens.append(next_token)
        decoded_tokens.append(decode([next_token]))

        x = jnp.concatenate((x[1:], jnp.array([next_token])))
    return decoded_tokens, tokens



import math
from dataclasses import dataclass
from typing import Literal


@dataclass
class MambaModelArgs:
    n_embd: int
    n_dims: int
    n_layers: int
    d_state: int = 16
    expand: int = 2
    dt_rank: int | str = "auto"
    d_conv: int = 4
    pad_vocab_size_multiple: int = 8
    conv_bias: bool = True
    bias: bool = False
    max_seq_len: int = 8
    d_inner: int | None = None
    use_in_projection_bias: bool = True
    use_conv_bias: bool = True
    use_out_proj_bias: bool = True
    ssm_use_delta_proj_bias: bool = False
    ssm_use_input_proj_bias: bool = False
    key_seed: int = 0

    def __post_init__(self):
        self.d_inner = int(self.expand * self.n_embd)

        if self.dt_rank == "auto":
            self.dt_rank = math.ceil(self.n_embd / self.d_state)

        if self.n_dims % self.pad_vocab_size_multiple != 0:
            self.n_dims += (
                self.pad_vocab_size_multiple
                - self.n_dims % self.pad_vocab_size_multiple
            )


@dataclass
class KiraModelArgs:
    n_dims: int
    n_embd: int
    n_layers: int
    max_seq_len: int
    num_heads: int
    num_query_heads: int
    num_kv_heads: int
    width_size: int
    depth: int
    key_seed: int
    p: float
    kv_interpolation_mode: Literal["average", "repeat"]


def get_mamba_args(
    n_embd: int,
    n_dims: int,
    n_layers: int,
    key_seed: int = 0,
    d_state: int = 16,
    expand: int = 2,
    dt_rank: int | str = "auto",
    d_conv: int = 4,
    pad_vocab_size_multiple: int = 8,
    conv_bias: bool = True,
    bias: bool = False,
    max_seq_len: int = 8,
) -> MambaModelArgs:
    return MambaModelArgs(
        n_embd=n_embd,
        n_dims=n_dims,
        n_layers=n_layers,
        d_state=d_state,
        expand=expand,
        dt_rank=dt_rank,
        d_conv=d_conv,
        pad_vocab_size_multiple=pad_vocab_size_multiple,
        conv_bias=conv_bias,
        bias=bias,
        max_seq_len=max_seq_len,
        key_seed=key_seed,
    )


def get_kira_args(
    n_dims: int,
    n_embd: int,
    n_layers: int,
    max_seq_len: int,
    num_heads: int,
    num_query_heads: int,
    num_kv_heads: int,
    width_size: int,
    depth: int,
    key_seed: int = 0,
    kv_interpolation_mode: Literal["average", "repeat"] = "average",
    p: float = 0.2,
) -> KiraModelArgs:
    return KiraModelArgs(
        n_dims=n_dims,
        n_embd=n_embd,
        n_layers=n_layers,
        max_seq_len=max_seq_len,
        num_heads=num_heads,
        num_query_heads=num_query_heads,
        num_kv_heads=num_kv_heads,
        width_size=width_size,
        depth=depth,
        p=p,
        key_seed=key_seed,
        kv_interpolation_mode=kv_interpolation_mode,
    )
